#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from ory.ext.exceptions import PermissionException
from ory.states.permission import Permission


class Pod:
    """
    A Pod is like a storage system.
    You can store any type of data in it, but the storage type
    must be the same for all stored data in the pod.
    """

    def __init__(self, name: str, storage: Any, permission: Permission, priority: int = 0) -> None:
        self.__name = name.lower()
        self.__storage = storage
        self.__permission = permission
        self.__priority = priority

    def get_name(self) -> str:
        """
        It gets the name of the current pod.
        :return: Name of the pod
        """

        return self.__name

    def get_storage(self) -> Any:
        """
        It gets the stored data of the current pod.
        :return: The pod storage data
        """

        return self.__storage

    def get_permission(self) -> Permission:
        """
        It gets the permission of the current pod.
        :return: Permission of the pod
        """

        return self.__permission

    def get_priority(self) -> int:
        """
        It gets the priority of the current pod.
        :return: The priority of the pod
        """

        return self.__priority

    def set_new_storage(self, storage: Any) -> None:
        """
        It overrides the current storage data with the new one.
        :param storage: The new storage data
        :return: None
        """

        if self.__permission != Permission.ALL:
            raise PermissionException("Cannot change instance of read only pod")

        self.__storage = storage

    def set_new_priority(self, priority: int) -> None:
        """
        It sets a new priority for the current pod.
        :param priority: The new priority
        :return: None
        """

        if self.__permission == Permission.READ_ONLY:
            raise PermissionException("Cannot change priority of read only pod")

        self.__priority = priority

    def set_new_name(self, name: str) -> None:
        """
        It sets the new the of the current pod.
        :param name: The new name
        :return: None
        """

        if self.__permission == Permission.READ_ONLY:
            raise PermissionException("Cannot change name of read only pod")

        self.__name = name



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from functools import wraps
from typing import Callable


def singleton(cls: Callable) -> Callable:
    """
    Decorate any class with this singleton func to make it as singleton class
    :param cls: The class to decorate
    :return: The singleton
    """

    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs) -> cls:
        """
        Get the instance of the singleton
        :param args: The arguments
        :param kwargs: The keyword arguments
        :return: The instance
        """

        instance = instances.setdefault(cls, cls(*args, **kwargs))
        return instance

    return get_instance



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from typing import Union

from ory.components.pod import Pod
from ory.ext.exceptions import PodAlreadyExistsException, PriorityAlreadyExistsException, PermissionException
from ory.states.permission import Permission


class Pool:
    """
    A pool is a collection of pods.
    It is used to group pods.
    """

    def __init__(self, name: str, permission: Permission, priority: int = 0) -> None:
        self.__name = name.lower()
        self.__permission = permission
        self.__priority = priority

        self.__pods = dict()

    def get_name(self) -> str:
        """
        It gets the name of the current pool.
        :return: The name of the pool
        """

        return self.__name

    def get_permission(self) -> Permission:
        """
        It gets the permission of the current pool.
        :return: The Permission of the pool
        """

        return self.__permission

    def get_priority(self) -> int:
        """
        It gets the priority of the current pool.
        :return: The priority of the pool
        """

        return self.__priority

    def register_pod(self, pod: Pod, override: bool = False) -> None:
        """
        This function registers a pod to the pool.
        If the pod already exists, you can override it.
        :param pod: The pod to register
        :param override: True if the pod should override an existing pod
        :return: None
        """

        if pod.get_name() in self.__pods and not override:
            raise PodAlreadyExistsException("Pod already registered")

        for registered_pod in self.__pods.values():
            if pod.get_priority() == registered_pod.get_priority() and pod.get_priority() != 0:
                raise PriorityAlreadyExistsException("Priority already exists")

            if pod.get_name() == registered_pod.get_name():
                raise PodAlreadyExistsException("Name already registered")

        if override and self.__permission == Permission.READ_ONLY:
            raise PermissionException("This pool is read only")

        self.__pods[pod.get_name()] = pod

    def register_pods(self, pods: list[Pod], override: bool = False) -> None:
        """
        This function registers multiple pods to the pool.
        If a pod already exists, you can override it.
        :param pods: The pods to register
        :param override: True if the pods should override an existing pod
        :return: None
        """

        for pod in pods:
            self.register_pod(pod, override)

    def unregister_pod(self, pod: Pod) -> None:
        """
        This function unregisters a pod from the pool.
        :param pod: The pod to unregister
        :return: None
        """

        if pod.get_name() not in self.__pods:
            raise PodAlreadyExistsException("Pod already registered")

        if self.__permission == Permission.READ_ONLY:
            raise PermissionException("This pool is read only")

        del self.__pods[pod.get_name()]

    def unregister_pods(self, pods: list[Pod]) -> None:
        """
        This function unregisters multiple pods from the pool.
        :param pods: The pods to unregister
        :return: None
        """

        for pod in pods:
            self.unregister_pod(pod)

    def get_all_pods(self) -> list:
        """
        This function gets all pods of the pool.
        :return: All pods of the pool
        """

        return list(self.__pods.values())

    def get_pod_by_name(self, name: str) -> Union[Pod, None]:
        """
        This function gets a pod by its name.
        :param name: The name of the pod.
        :return: If the pod exists, it returns the pod. Otherwise, it returns None.
        """

        for pod in self.__pods.values():
            if pod.get_name() == name.lower():
                return pod

        return None

    def get_pod_by_priority(self, priority: int) -> Union[list[Pod], Pod, None]:
        """
        This function gets a pod by its priority.
        :param priority: The priority of the pod.
        :return: If the priority is zero, it returns all pods with the priority zero. Otherwise, it returns the pod with the given priority or None if it does not exist.
        """

        if priority == 0:
            return [pod for pod in self.__pods.values() if pod.get_priority() == 0]

        for pod in self.__pods.values():
            if pod.get_priority() == priority:
                return pod

        return None

    def get_standard_pods(self) -> list[Pod]:
        """
        This function gets all pods with the priority zero.
        :return: All pods with the priority zero
        """

        return [pod for pod in self.__pods.values() if pod.get_priority() == 0]



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from ory.ext.exceptions import PermissionException
from ory.states.permission import Permission


class AsyncPod:
    """
    The AsyncPod is like the normal Pod, but it is async.
    You can also store any type of data in it, but the storage type
    must be the same for all stored data in the pod too.
    """

    def __init__(self, name: str, storage: Any, permission: Permission, priority: int = 0) -> None:
        self.__name = name.lower()
        self.__storage = storage
        self.__permission = permission
        self.__priority = priority

    async def get_name(self) -> str:
        """
        It gets the name of the current pod.
        :return: Name of the pod
        """

        return self.__name

    async def get_storage(self) -> Any:
        """
        It gets the stored data of the current pod.
        :return: The pod storage data
        """

        return self.__storage

    async def get_permission(self) -> Permission:
        """
        It gets the permission of the current pod.
        :return: Permission of the pod
        """

        return self.__permission

    async def get_priority(self) -> int:
        """
        It gets the priority of the current pod.
        :return: The priority of the pod
        """

        return self.__priority

    async def set_new_storage(self, storage: Any) -> None:
        """
        It overrides the current storage data with the new one.
        :param storage: The new storage data
        :return: None
        """

        if self.__permission != Permission.ALL:
            raise PermissionException("Cannot change instance of read only pod")

        self.__storage = storage

    async def set_new_priority(self, priority: int) -> None:
        """
        It sets a new priority for the current pod.
        :param priority: The new priority
        :return: None
        """

        if self.__permission == Permission.READ_ONLY:
            raise PermissionException("Cannot change priority of read only pod")

        self.__priority = priority

    async def set_new_name(self, name: str) -> None:
        """
        It sets the new the of the current pod.
        :param name: The new name
        :return: None
        """

        if self.__permission == Permission.READ_ONLY:
            raise PermissionException("Cannot change name of read only pod")

        self.__name = name



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from typing import Union
from ory.states.permission import Permission

from ory.components.pool import Pool
from ory.ext.exceptions import PoolAlreadyExistsException, PriorityAlreadyExistsException
from ory.registry.pool_registry import PoolRegistry


class RegistryHandler:
    """
    This is used to communicate with the pool registry.
    It handels all requests and make the pool public in the entire application.
    """

    @staticmethod
    def create_new_pool(pool: Pool, override: bool = False) -> None:
        """
        This function is used to create a new pool.
        It adds the pool to the pool registry.

        Attributes:
        ------------
        :param pool: The pool to create
        :param override: If the pool should be overriden if it already exists
        :return: None
        """

        if pool.get_name() in PoolRegistry.pool_storage and not override:
            raise PoolAlreadyExistsException("Service pool already exists")

        for registered_pool in PoolRegistry.pool_storage.values():
            if registered_pool.get_priority() == pool.get_priority() and pool.get_priority() != 0:
                raise PriorityAlreadyExistsException("Service pool with same priority already exists")

        if pool.get_name() in PoolRegistry.pool_storage and not override:
            raise PoolAlreadyExistsException("Service pool already exists")

        PoolRegistry.pool_storage[pool.get_name()] = pool

    @staticmethod
    def create_new_pools(pools: list[Pool], override: bool = False) -> None:
        """
        This function is used to create multiple pools.
        :param pools: The pools to create
        :param override: If the pools should be overriden if they already exist
        :return: None
        """

        for pool in pools:
            RegistryHandler.create_new_pool(pool, override)

    @staticmethod
    def delete_pool(pool: Pool) -> None:
        """
        This function is used to delete a pool.
        It removes the pool from the pool registry.
        :param pool: The pool to delete
        :return: None
        """

        if pool.get_permission() != Permission.ALL:
            raise PermissionError("Missing permission to delete pool")

        del PoolRegistry.pool_storage[pool.get_name()]

    @staticmethod
    def get_all_pools() -> list:
        """
        This function is used to get all pools.
        :return: All pools
        """

        return [pool for pool in PoolRegistry.pool_storage.values()]

    @staticmethod
    def get_pool_by_name(name: str) -> Union[Pool, None]:
        """
        This function is used to get a pool by its name.
        :param name: The name of the pool.
        :return: If the pool exists, it will return the pool. If not, it will return None.
        """

        for pool in PoolRegistry.pool_storage.values():
            if pool.get_name() == name.lower():
                return pool

        return None

    @staticmethod
    def get_pool_by_priority(priority: int) -> Union[list[Pool], Pool, None]:
        """
        This function is used to get a pool by its priority.
        :param priority: The priority of the pool.
        :return: If the priority is zero, it will return all pools with the priority zero. If not, it will return the pool with the given priority or None if it does not exist.
        """

        if priority == 0:
            return [pool for pool in PoolRegistry.pool_storage.values() if pool.get_priority() == 0]

        for pool in PoolRegistry.pool_storage.values():
            if pool.get_priority() == priority:
                return pool

        return None

    @staticmethod
    def get_standard_pools() -> list[Pool]:
        """
        This function is used to get all standard pools.
        Standard pools are pools with the priority zero.
        :return: All standard pools
        """

        return [pool for pool in PoolRegistry.pool_storage.values() if pool.get_priority() == 0]



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

from ory.ext.singleton import Singleton


class PoolRegistry(Singleton):
    """
    The pool registry is used to store all the pools. Normally, you should not use this class directly.
    It is responsible for the pool management and that the pools are globally accessible.
    """

    pool_storage = dict()



#  Copyright 2023 Nyria
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

class Singleton:
    """
    That class represents a singleton. Any class that inherits from this class will be a singleton.
    """

    _instance = None

    def __new__(cls, *args, **kwargs) -> object:
        """
        Creates a new instance of the class if it doesn't exist.
        """

        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)

        return cls._instance
